/*
Smart Array Class
3 January 2005

CArrayIndex<T> - A generic dynamic array class
	T must have a constructor with no arguments
		this constructor doesn't have to initialise anything properly
	T must support operator =
TArray - Uses the reference counting on TArray to support copy-on-write
CFastArrayIndex<T> - A generic dynamic array class
	T must have a constructor with no arguments
	Uses memcpy() to copy internal arrays (instead of operator =)
	Probably only works for primitive types or classes/structs with no pointer members
TFastArray - Uses the reference counting on TFastArray for fast copying when passing by value
*/

namespace tlib
{

template <typename T> void ArrayIndexSort(T* data, int index0, int index1,
		bool (*sortFunction)(T&, T&))
{
	int NumItems = index1 - index0;
	if (NumItems < 2)
		return;
	int Pivot = index0 + (NumItems - 1) / 2;
	int NewPivot = index0;
	TSwap(data[index0], data[Pivot]);
	for (int i = index0 + 1; i < index1; i++)
	{
		if (sortFunction(data[i], data[index0]))
		{
			NewPivot++;
			if (i != NewPivot)
				TSwap(data[i], data[NewPivot]);
		}
	}
	if (NewPivot == index0)
		ArrayIndexSort(data, index0 + 1, index1, sortFunction);
	else
	{
		TSwap(data[NewPivot], data[index0]);
		ArrayIndexSort(data, index0, NewPivot, sortFunction);
		ArrayIndexSort(data, NewPivot + 1, index1, sortFunction);
	}
}

template <typename T> class CArrayIndex
{
public:
	CArrayIndex()
	{
		FReferenceCount = 1;

		FData = NULL;
		FMaxItems = 0;
		FNumItems = 0;
	}
	CArrayIndex(const T* data, int numItems, int offset=0)
	{
		FReferenceCount = 1;

		FData = NULL;
		FMaxItems = 0;
		FNumItems = numItems;
		if (FNumItems != 0)
		{
			AllocArray(FNumItems);
			for (int i = 0; i < FNumItems; i++)
				FData[i] = data[i + offset];
		}
	}
	CArrayIndex(const T& item, const CArrayIndex<T>* array)
	{
		FReferenceCount = 1;

		FData = NULL;
		FMaxItems = 0;

		AllocArray(array->FNumItems + 1);
		memcpy(FData, array->FData, array->FNumItems * sizeof(T));
		FData[array->FNumItems] = item;
		FNumItems = array->FNumItems + 1;
	}
	CArrayIndex(const CArrayIndex<T>* array)
	{
		FReferenceCount = 1;

		FData = NULL;
		FMaxItems = 0;
		FNumItems = array->FNumItems;
		if (FNumItems)
		{
			AllocArray(FNumItems);
			for (int i = 0; i < FNumItems; i++)
				FData[i] = array->FData[i];
		}
	}
	~CArrayIndex()
	{
		if (FData)
			delete [] FData;
	}

	CArrayIndex<T>* AddRef()
	{
		FReferenceCount++;
		return this;
	}
	void Release()
	{
		if (FReferenceCount < 0)
			throw TException(L"Invalid array");
		if (!--FReferenceCount)
			delete this;
	}

	void Reset()
	{
		if (FData)
			delete [] FData;
		FNumItems = 0;
		FData = NULL;
		FMaxItems = 0;
	}

	void Add(const T& item)
	{
		AllocArray(FNumItems + 1);
		FData[FNumItems++] = item;
	}
	void Insert(const T& item, int index)
	{
		if (index == FNumItems)
		{
			Add(item);
			return;
		}
		if (index < 0 || index > FNumItems)
			throw TException(L"Item index out of bounds");
		AllocArray(FNumItems + 1);
		for (int i = FNumItems; i > index; i--)
			FData[i] = FData[i - 1];
		FData[index] = item;
		FNumItems++;
	}

	void Remove(int index)
	{
		if (index < 0 || index >= FNumItems)
			throw TException(L"Item index out of bounds");
		FNumItems--;
		for (int i = index; i < FNumItems; i++)
			FData[i] = FData[i + 1];
	}
	void Remove(int startIndex, int endIndex)
	{
		if (endIndex < startIndex)
			throw TException(L"Invalid arguments");
		if (startIndex < 0 || startIndex >= FNumItems)
			throw TException(L"Start index out of bounds");
		if (endIndex < 0 || endIndex >= FNumItems)
			throw TException(L"End index out of bounds");

		int NumMoveItems = FNumItems - (endIndex + 1);
		for (int i = 0; i < NumMoveItems; i++)
			FData[i + startIndex] = FData[i + endIndex + 1];
		FNumItems -= (endIndex - startIndex + 1);
	}
	void RemoveLength(int startIndex, int length)
	{
		if (length <= 0)
			throw TException(L"Invalid length");
		if (startIndex < 0 || startIndex >= FNumItems)
			throw TException(L"Start index out of bounds");

		int NumMoveItems = FNumItems - (startIndex + length);
		for (int i = 0; i < NumMoveItems; i++)
			FData[i + startIndex] = FData[i + startIndex + length];
		FNumItems -= length;
	}
	void RemoveItem(const T& item)
	{
		int Index = IndexOf(item);
		if (Index == -1)
			throw TException(L"Item is not in array");
		Remove(Index);
	}

	int IndexOf(const T& item)
	{
		for (int i = 0; i < FNumItems; i++)
			if (FData[i] == item)
				return i;
		return -1;
	}
	bool Contains(const T& item)
	{
		for (int i = 0; i < FNumItems; i++)
			if (FData[i] == item)
				return true;
		return false;
	}

	void Clear()
	{
		FNumItems = 0;
	}

	void Sort(bool (*sortFunction)(T&, T&))
	{
		ArrayIndexSort(FData, 0, FNumItems, sortFunction);
	}

	PROPERTY_V(NumItems, int, FNumItems)
	T& Item(int i) const
	{
		if (i < 0 || i >= FNumItems)
			throw TException(L"Item index out of bounds");
		return FData[i];
	}
	PROPERTY_F(Last, T&)

	PROPERTY_F(Shared, bool)

	PROPERTY_C(Data, T*, FData)

private:
	int FReferenceCount;
	bool _GetShared() const { return FReferenceCount != 1; }

	int FNumItems;
	T *FData;
	int FMaxItems;

	void AllocArray(int minItems)
	{
		if (!minItems)
			throw TException(L"minItems == 0");

		if (FMaxItems >= minItems)
			return;
		FMaxItems = TMax(TMax(FMaxItems + FMaxItems / 2, minItems), 8);

		T *OldData = FData;
		FData = new T [FMaxItems];
		if (OldData)
		{
			for (int i = 0; i < FNumItems; i++)
				FData[i] = OldData[i];
			delete [] OldData;
		}
	}

	T& _GetLast() const
	{
		if (FNumItems == 0)
			throw TException(L"Array is empty");
		return FData[FNumItems - 1];
	}
};



template <typename T> class CFastArrayIndex
{
public:
	CFastArrayIndex()
	{
		FReferenceCount = 1;

		FData = NULL;
		FMaxItems = 0;
		FNumItems = 0;
	}
	CFastArrayIndex(const T* data, int numItems, int offset=0)
	{
		FReferenceCount = 1;

		FData = NULL;
		FMaxItems = 0;
		FNumItems = numItems;
		if (FNumItems != 0)
		{
			AllocArray(FNumItems);
			memcpy(FData, data + offset, FNumItems * sizeof(T));
		}
	}
	CFastArrayIndex(const T& item, const CFastArrayIndex<T>* array)
	{
		FReferenceCount = 1;

		FData = NULL;
		FMaxItems = 0;

		AllocArray(array->FNumItems + 1);
		memcpy(FData, array->FData, array->FNumItems * sizeof(T));
		FData[array->FNumItems] = item;
		FNumItems = array->FNumItems + 1;
	}
	CFastArrayIndex(const CFastArrayIndex<T>* array)
	{
		FReferenceCount = 1;

		FData = NULL;
		FMaxItems = 0;
		FNumItems = array->FNumItems;
		if (FNumItems)
		{
			AllocArray(FNumItems);
			memcpy(FData, array->FData, FNumItems * sizeof(T));
		}
	}
	~CFastArrayIndex()
	{
		if (FData)
			delete [] FData;
	}

	CFastArrayIndex<T>* AddRef()
	{
		FReferenceCount++;
		return this;
	}
	void Release()
	{
		if (FReferenceCount < 0)
			throw TException(L"Invalid array");
		if (!--FReferenceCount)
			delete this;
	}

	void Reset()
	{
		if (FData)
			delete [] FData;
		FNumItems = 0;
		FData = NULL;
		FMaxItems = 0;
	}

	void Add(const T& item)
	{
		AllocArray(FNumItems + 1);
		FData[FNumItems++] = item;
	}
	void Insert(const T& item, int index)
	{
		if (index == FNumItems)
		{
			Add(item);
			return;
		}
		if (index < 0 || index > FNumItems)
			throw TException(L"Item index out of bounds");
		AllocArray(FNumItems + 1);
		memmove(FData + index + 1, FData + index, (FNumItems - index) * sizeof(T));
		FData[index] = item;
		FNumItems++;
	}

	void Remove(int index)
	{
		if (index < 0 || index >= FNumItems)
			throw TException(L"Item index out of bounds");
		memcpy(FData + index, FData + index + 1, (FNumItems - 1 - index) * sizeof(T));
		FNumItems--;
	}
	void Remove(int startIndex, int endIndex)
	{
		if (endIndex < startIndex)
			throw TException(L"Invalid arguments");
		if (startIndex < 0 || startIndex >= FNumItems)
			throw TException(L"Start index out of bounds");
		if (endIndex < 0 || endIndex >= FNumItems)
			throw TException(L"End index out of bounds");

		memcpy(FData + startIndex, FData + endIndex + 1, (FNumItems - 1 - endIndex) * sizeof(T));
		FNumItems -= (endIndex - startIndex + 1);
	}
	void RemoveLength(int startIndex, int length)
	{
		if (length <= 0)
			throw TException(L"Invalid length");
		if (startIndex < 0 || startIndex >= FNumItems)
			throw TException(L"Start index out of bounds");

		memcpy(FData + startIndex, FData + startIndex + length, (FNumItems - (startIndex + length)) * sizeof(T));
		FNumItems -= length;
	}
	void RemoveItem(const T& item)
	{
		int Index = IndexOf(item);
		if (Index == -1)
			throw TException(L"Item is not in array");
		Remove(Index);
	}

	int IndexOf(const T& item)
	{
		for (int i = 0; i < FNumItems; i++)
			if (FData[i] == item)
				return i;
		return -1;
	}
	bool Contains(const T& item)
	{
		for (int i = 0; i < FNumItems; i++)
			if (FData[i] == item)
				return true;
		return false;
	}

	void Clear()
	{
		FNumItems = 0;
	}

	void Sort(bool (*sortFunction)(T&, T&))
	{
		ArrayIndexSort(FData, 0, FNumItems, sortFunction);
	}

	PROPERTY_V(NumItems, int, FNumItems)
	T& Item(int i) const
	{
		if (i < 0 || i >= FNumItems)
			throw TException(L"Item index out of bounds");
		return FData[i];
	}

	PROPERTY_F(Last, T&)

	PROPERTY_F(Shared, bool)

	PROPERTY_C(Data, T*, FData)

private:
	int FReferenceCount;
	bool _GetShared() const { return FReferenceCount != 1; }

	int FNumItems;
	T *FData;
	int FMaxItems;

	void AllocArray(int minItems)
	{
		if (!minItems)
			throw TException(L"minItems == 0");

		if (FMaxItems >= minItems)
			return;
		FMaxItems = TMax(TMax(FMaxItems + FMaxItems / 2, minItems), 8);

		T *OldData = FData;
		FData = new T [FMaxItems];
		if (OldData)
		{
			memcpy(FData, OldData, FNumItems * sizeof(T));
			delete [] OldData;
		}
	}

	T& _GetLast() const
	{
		if (FNumItems == 0)
			throw TException(L"Array is empty");
		return FData[FNumItems - 1];
	}
};



template <typename A, typename T> class TSmartArray
{
public:
	TSmartArray()
	{
		FArray = new A();
	}
	TSmartArray(const TSmartArray<A, T>& array)
	{
		FArray = array.FArray->AddRef();
	}
	~TSmartArray()
	{
		FArray->Release();
	}

	TSmartArray& operator =(const TSmartArray<A, T>& array)
	{
		FArray->Release();
		FArray = array.FArray->AddRef();
		return *this;
	}

	void Add(const T& item)
	{
		UniquifyArray();
		FArray->Add(item);
	}
	void Insert(const T& item, int index)
	{
		UniquifyArray();
		FArray->Insert(item, index);
	}

	void Remove(int index)
	{
		UniquifyArray();
		FArray->Remove(index);
	}
	void Remove(int startIndex, int endIndex)
	{
		UniquifyArray();
		FArray->Remove(startIndex, endIndex);
	}
	void RemoveLength(int startIndex, int length)
	{
		UniquifyArray();
		FArray->RemoveLength(startIndex, length);
	}
	void RemoveItem(const T& item)
	{
		UniquifyArray();
		FArray->RemoveItem(item);
	}

	int IndexOf(const T& item)
	{
		return FArray->IndexOf(item);
	}
	bool Contains(const T& item)
	{
		return FArray->Contains(item);
	}
	T& Find(const T& item)
	{
		int Index = FArray->IndexOf(item);
		if (Index == -1)
			throw TException(L"Item is not in array");
		return GetItem(Index);
	}

	void Reset()
	{
		UniquifyArray();
		FArray->Reset();
	}

	void Clear()
	{
		UniquifyArray();
		FArray->Clear();
	}

	void Sort(bool (*sortFunction)(T&, T&))
	{
		UniquifyArray();
		FArray->Sort(sortFunction);
	}

	void SetData(const T* data, int numItems, int offset=0)
	{
		FArray->Release();
		FArray = new A(data, numItems, offset);
	}

	PROPERTY_F(NumItems, int)
	T& operator [](int i) const
	{
		return GetItem(i);
	}

	PROPERTY_F(Last, T&)

	PROPERTY_C(Array, A*, FArray)

private:
	A *FArray;
	void UniquifyArray()
	{
		if (FArray->Shared)
		{
			A *NewArray = new A(FArray);
			FArray->Release();
			FArray = NewArray;
		}
	}

	int _GetNumItems() const { return FArray->NumItems; }

	T& _GetLast() const
	{
		if (FArray->NumItems == 0)
			throw TException(L"Array is empty");
		return FArray->Item(FArray->NumItems - 1);
	}

	T& GetItem(int i) const { return FArray->Item(i); }
};

template <typename T> class TArray : public TSmartArray<CArrayIndex<T>, T> { };
template <typename T> class TFastArray : public TSmartArray<CFastArrayIndex<T>, T> { };

}   // namespace tlib