/*
TStringMap class
4 September 2007

Implements a hashmap with string keys.
*/

namespace tlib
{

template <typename V> struct SStringPair
{
	SStringPair() { }
	SStringPair(TString key, V value)
			: Key(key), Value(value), Next(NULL)
	{
	}

	TString Key;
	V Value;
	SStringPair<V> *Next;
};



template <typename V> class CStringMapIndex
{
public:
	CStringMapIndex()
	{
		FReferenceCount = 1;
		FCapacity = 8;
		FRehashCapacity = FCapacity * 0.8;
		FData = new SStringPair<V>* [FCapacity];
		memset(FData, 0, FCapacity * sizeof(SStringPair<V>*));
		FNumItems = 0;
	}



	CStringMapIndex(const CStringMapIndex* index)
	{
		FReferenceCount = 1;
	}



	~CStringMapIndex()
	{
		for (int i = 0; i < FCapacity; i++)
		{
			SStringPair<V> *e = FData[i];
			while (e != NULL)
			{
				SStringPair<V> *Next = e->Next;
				delete e;
				e = Next;
			}
		}
		delete [] FData;
	}



	CStringMapIndex* AddRef()
	{
		FReferenceCount++;
		return this;
	}



	void Release()
	{
		if (--FReferenceCount == 0)
			delete this;
	}



	void Add(TString key, V value)
	{
		if (ContainsKey(key))
			throw TException(L"Key already exists");
		Set(key, value);
	}

	void Set(TString key, V value)
	{
		unsigned int Index = Hash(key) % FCapacity;
		SStringPair<V> *LastEntry = NULL;
		for (SStringPair<V>* e = FData[Index]; e != NULL; e = e->Next)
		{
			LastEntry = e;
			if (e->Key == key)
			{
				e->Value = value;
				return;
			}
		}
		SStringPair<V> *Entry = new SStringPair<V>(key, value);
		if (LastEntry == NULL)
			FData[Index] = Entry;
		else
			LastEntry->Next = Entry;
		if (FNumItems == FRehashCapacity)
			Rehash();
		FNumItems++;
	}



	void Remove(TString key)
	{
		unsigned int Index = Hash(key) % FCapacity;
		SStringPair<V> *Last = NULL;
		for (SStringPair<V>* e = FData[Index]; e != NULL; e = e->Next)
		{
			if (e->Key == key)
			{
				if (Last == NULL)
					FData[Index] = e->Next;
				else
					Last->Next = e->Next;
				delete e;
				FNumItems--;
				// TODO: rehash if NumEntries is below a threshold
				return;
			}
			Last = e;
		}
		throw TException(L"Key does not exist");
	}



	void RemoveIfExists(TString key)
	{
		if (ContainsKey(key))
			Remove(key);
	}



	V& Lookup(TString key) const
	{
		unsigned int Index = Hash(key) % FCapacity;
		for (SStringPair<V>* e = FData[Index]; e != NULL; e = e->Next)
		{
			if (e->Key == key)
				return e->Value;
		}
		throw TException(L"Key does not exist");
	}



	bool ContainsKey(TString key) const
	{
		unsigned int Index = Hash(key) % FCapacity;
		for (SStringPair<V>* e = FData[Index]; e != NULL; e = e->Next)
		{
			if (e->Key == key)
				return true;
		}
		return false;
	}



	static bool StringMapIndexEquals(const CStringMapIndex<V>* a, const CStringMapIndex<V>* b)
	{
		if (a->FNumItems != b->FNumItems)
			return false;
		TArray<TString> Keys = a->GetKeyList();
		TArray<V> Values = a->GetValueList();
		for (int i = 0; i < Keys.NumItems; i++)
		{
			if (!b->ContainsKey(Keys[i]))
				return false;
			if (!(Values[i] == b->Lookup(Keys[i])))
				return false;
		}
		return true;
	}



	TArray<TString> GetKeyList() const
	{
		TArray<TString> KeyList;
		for (int i = 0; i < FCapacity; i++)
		{
			for (SStringPair<V> *e = FData[i]; e != NULL; e = e->Next)
				KeyList.Add(e->Key);
		}
		return KeyList;
	}



	TArray<V> GetValueList() const
	{
		TArray<V> ValueList;
		for (int i = 0; i < FCapacity; i++)
		{
			for (SStringPair<V> *e = FData[i]; e != NULL; e = e->Next)
				ValueList.Add(e->Value);
		}
		return ValueList;
	}



	PROPERTY_F(Shared, bool)

private:
	int FReferenceCount;
	bool _GetShared() const { return FReferenceCount != 1; }
	int FCapacity;
	int FRehashCapacity;
	SStringPair<V> **FData;
	int FNumItems;



	unsigned int Hash(tlib::TString s) const
	{
		const TByte *Data = (const TByte*)s.Str;
		int Size = s.Size * sizeof(wchar_t);
		unsigned int Hash = 0;
		for (int i = 0; i < Size; i++)
		{
			Hash *= 7;
			Hash += Data[i] + 1;
		}
		return Hash;
	}



	/**
	Increases the capacity of the table and rebuilds it.
	*/
	void Rehash()
	{
		SStringPair<V> **OldData = FData;
		int OldCapacity = FCapacity;
		FCapacity += FCapacity;
		FData = new SStringPair<V>* [FCapacity];
		memset(FData, 0, FCapacity * sizeof(SStringPair<V>*));

		for (int i = 0; i < OldCapacity; i++)
		{
			SStringPair<V> *e = OldData[i];
			while (e != NULL)
			{
				SStringPair<V> *Next = e->Next;
				e->Next = NULL;
				int Index = Hash(e->Key) % FCapacity;
				SStringPair<V> *Last = FData[Index];
				if (Last == NULL)
					FData[Index] = e;
				else
				{
					while (Last->Next != NULL)
						Last = Last->Next;
					Last->Next = e;
				}

				e = Next;
			}
		}
		delete [] OldData;

		FRehashCapacity = FCapacity * 0.8;
	}
};



template <typename V> class TStringMap
{
public:
	TStringMap()
	{
		FIndex = new CStringMapIndex<V>();
	}
	~TStringMap()
	{
		FIndex->Release();
	}

	TStringMap(const TStringMap<V>& rhs)
	{
		FIndex = rhs.FIndex->AddRef();
	}

	TStringMap& operator =(const TStringMap<V>& rhs)
	{
		FIndex->Release();
		FIndex = rhs.FIndex->AddRef();
		return *this;
	}

	bool operator ==(const TStringMap<V>& rhs) const
	{
		return CStringMapIndex<V>::StringMapIndexEquals(FIndex, rhs.FIndex);
	}

	bool operator !=(const TStringMap<V>& rhs) const
	{
		return !CStringMapIndex<V>::StringMapIndexEquals(FIndex, rhs.FIndex);
	}

	void Add(TString key, V value)
	{
		if (FIndex->ContainsKey(key))
			throw TException(L"Key already exists");
		Set(key, value);
	}

	void Set(TString key, V value)
	{
		Uniquify();
		FIndex->Set(key, value);
	}

	void Remove(TString key)
	{
		Uniquify();
		FIndex->Remove(key);
	}

	void RemoveIfExists(TString key)
	{
		Uniquify();
		FIndex->RemoveIfExists(key);
	}

	V& operator [](TString key) const
	{
		return FIndex->Lookup(key);
	}

	bool ContainsKey(TString key) const
	{
		return FIndex->ContainsKey(key);
	}

	PROPERTY_F(KeyList, TArray<TString>)
	PROPERTY_F(ValueList, TArray<V>)

private:
	CStringMapIndex<V> *FIndex;

	void Uniquify()
	{
		if (FIndex->Shared)
		{
			FIndex->Release();
			FIndex = new CStringMapIndex<V>(FIndex);
		}
	}

	TArray<TString> _GetKeyList() const
	{
		return FIndex->GetKeyList();
	}

	TArray<V> _GetValueList() const
	{
		return FIndex->GetValueList();
	}
};

}